use crate::AppState;
use actix_web::{
    error::{ErrorBadRequest, ErrorInternalServerError},
    web, HttpResponse, Responder,
};
use chrono::Utc;
use entity::{node, node_group, sensor_data};
use sea_orm::{entity::*, query::*, ActiveModelTrait, ActiveValue, DbBackend, EntityTrait};
use serde::{Deserialize, Serialize};
use uuid::Uuid;

#[derive(Serialize)]
struct NodeWithSensorData {
    node: node::Model,
    sensor_data: Vec<sensor_data::Model>,
}

#[derive(Deserialize)]
pub struct CreateGroupWithoutId {
    name: String,
}

#[derive(Deserialize)]
struct CreateNodeRequest {
    id: u64,
    group: uuid::Uuid,
}

#[derive(Serialize)]
struct GroupWithNode {
    #[serde(flatten)]
    group: node_group::Model,
    node: Vec<node::Model>,
}

pub async fn get_nodes(state: web::Data<AppState>) -> actix_web::Result<impl Responder> {
    let db = &state.db;

    let result = node_group::Entity::find()
        .find_with_related(entity::prelude::Node)
        .all(db)
        .await
        .map_err(ErrorInternalServerError)?
        .into_iter()
        .map(|(group, node)| GroupWithNode { group, node })
        .collect::<Vec<_>>();

    Ok(web::Json(result))
}

pub async fn get_data(state: web::Data<AppState>) -> actix_web::Result<impl Responder> {
    let db = &state.db;

    let nodes = node::Entity::find()
        .all(db)
        .await
        .map_err(ErrorInternalServerError)?;

    let now = Utc::now();
    let one_hour_ago = now - chrono::Duration::hours(1);

    let mut result: Vec<NodeWithSensorData> = Vec::new();
    for node in nodes {
        let node_id = node.id.clone();
        let sensor_data = sensor_data::Entity::find()
            .filter(sensor_data::Column::NodeId.eq(node_id))
            .filter(sensor_data::Column::Timestamp.gt(one_hour_ago))
            .all(db)
            .await
            .map_err(ErrorInternalServerError)?;

        result.push(NodeWithSensorData { node, sensor_data });
    }

    Ok(web::Json(result))
}

pub async fn create_group(
    state: web::Data<AppState>,
    group: web::Json<CreateGroupWithoutId>,
) -> actix_web::Result<impl Responder> {
    let db = &state.db;

    let group = group.into_inner();

    let group = entity::node_group::ActiveModel {
        id: ActiveValue::NotSet,
        name: ActiveValue::Set(group.name),
    };

    let result = group.insert(db).await.map_err(ErrorInternalServerError)?;

    Ok(web::Json(result))
}

pub async fn create_node(
    state: web::Data<AppState>,
    node_request: web::Json<CreateNodeRequest>,
) -> actix_web::Result<impl Responder> {
    let db = &state.db;

    let node = node_request.into_inner();

    let node = entity::node::ActiveModel {
        id: ActiveValue::Set(node.id),
        group: ActiveValue::Set(node_request.group),
    };

    let result = node.insert(db).await.map_err(ErrorInternalServerError)?;

    Ok(web::Json(result))
}

/*
pub async fn delete_node(
    state: web::Data<AppState>,
    path: web::Path<Uuid>,
) -> actix_web::Result<impl Responder> {
    let id = path.into_inner();

    let db = &state.db;

    entity::node::Entity::delete_by_id(id)
        .exec(db)
        .await
        .map_err(ErrorInternalServerError)?;

    Ok(HttpResponse::Ok().finish())
}
*/

fn mac_to_i32(mac: &str) -> Result<i32, std::num::ParseIntError> {
    // Remove non-hexadecimal characters
    let sanitized_mac: String = mac.chars().filter(|c| c.is_digit(16)).collect();

    // Parse the sanitized string as a hexadecimal number
    i32::from_str_radix(&sanitized_mac, 16)
}
